import pytest

from hstrat import hstrat


@pytest.mark.filterwarnings(
    "ignore:Animation was deleted without rendering anything."
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.geom_seq_nth_root_algo.Policy(4),
    ],
)
def test_one(policy):
    hstrat.stratum_retention_animate(
        policy,
        10,
        draw_extant_history=False,
        draw_extinct_history=False,
        draw_extinct_placeholders=True,
    ).to_html5_video(
        embed_limit=0
    )  # silence mpl unused animation warning
    hstrat.stratum_retention_animate(
        policy,
        10,
        draw_extant_history=True,
        draw_extinct_history=False,
        draw_extinct_placeholders=True,
    ).to_html5_video(
        embed_limit=0
    )  # silence mpl unused animation warning
    hstrat.stratum_retention_animate(policy, 10).to_html5_video(
        embed_limit=0
    )  # silence mpl unused animation warning


@pytest.mark.heavy
def test_doc_animations(docplot_policy):
    hstrat.stratum_retention_animate(
        docplot_policy,
        256,
        save_as="gif",
    )


@pytest.mark.heavy
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.depth_proportional_resolution_tapered_algo.Policy(
            parameterizer=hstrat.PropertyAtMostParameterizer(
                target_value=10,
                policy_evaluator=hstrat.NumStrataRetainedExactEvaluator(
                    at_num_strata_deposited=256,
                ),
                param_lower_bound=1,
                param_upper_bound=1024,
            )
        )
    ],
)
def test_more_doc_animations(policy):

    hstrat.stratum_retention_animate(
        policy,
        256,
        draw_extant_history=False,
        draw_extinct_history=False,
        draw_extinct_placeholders=True,
        save_as="gif",
    )
    hstrat.stratum_retention_animate(
        policy,
        256,
        draw_extant_history=True,
        draw_extinct_history=False,
        draw_extinct_placeholders=True,
        save_as="gif",
    )
    hstrat.stratum_retention_animate(
        policy,
        256,
        save_as="gif",
    )



import typing

from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ._calc_rank_of_parity_segue_between import (
    calc_rank_of_parity_segue_between,
)


def calc_rank_of_last_retained_commonality_between(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find rank of strata commonality before first strata disparity.

    Implementation detail with general-case implementation.
    """
    return calc_rank_of_parity_segue_between(
        first,
        second,
        confidence_level,
        None,
    )[0]



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_find_root_ids,
    alifestd_make_empty,
    alifestd_mark_node_depth_asexual,
    alifestd_unfurl_traversal_postorder_asexual,
    is_nondecreasing,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_unfurl_traversal_postorder_asexual(phylogeny_df)

    assert original.equals(phylogeny_df)

    assert len(result) == len(phylogeny_df)
    assert set(result) == set(phylogeny_df["id"])

    reference_df = alifestd_mark_node_depth_asexual(phylogeny_df)
    node_depths = dict(
        zip(
            reference_df["id"],
            reference_df["node_depth"],
        ),
    )
    node_depths = [node_depths[id_] for id_ in reversed(result)]
    assert is_nondecreasing(node_depths)

    assert result[-1] in alifestd_find_root_ids(phylogeny_df)
    assert result[0] in alifestd_find_leaf_ids(phylogeny_df)


def test_empty():
    res = alifestd_unfurl_traversal_postorder_asexual(alifestd_make_empty())
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_postorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result.tolist() == [2, 1, 0]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_postorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result.tolist() == [3, 2, 0, 1]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [4, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[4]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_postorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result.tolist() == [3, 2, 4, 0]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[0]", "[0]", "[1]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_postorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )

    assert result.tolist() == [3, 2, 1, 0]

    if not mutate:
        assert original_df.equals(phylogeny_df)



import numpy as np
import pandas as pd


def alifestd_find_root_ids(phylogeny_df: pd.DataFrame) -> np.ndarray:  # int
    """What ids have an empty `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    root_mask = (
        (phylogeny_df["ancestor_id"] == phylogeny_df["id"])
        if "ancestor_id" in phylogeny_df
        else (
            phylogeny_df["ancestor_list"]
            .astype(str)
            .str.lower()
            .isin(("[none]", "[]"))
        )
    )

    return phylogeny_df.loc[root_mask, "id"].to_numpy().copy()



import typing

import numpy as np
import pandas as pd

from .._auxiliary_lib import (
    get_hstrat_version,
    log_once_in_a_row,
    numpy_fromiter_polyfill,
)
from ..frozen_instrumentation import HereditaryStratigraphicSpecimen
from ._impl import policy_from_record
from ._unpack_differentiae_str import unpack_differentiae_str


def specimen_from_records(
    records: typing.Dict,
) -> HereditaryStratigraphicSpecimen:
    """Deserialize a `HereditaryStratigraphicSpecimen` from a dict composed of
    builtin data types.

    See Also
    --------
    HereditaryStratigraphicSpecimen
        Postprocessing representation of the differentia retained
        by an extant HereditaryStratigraphicColumn, indexed by deposition rank.
    """
    if "deposition_ranks" in records:
        raise NotImplementedError

    if get_hstrat_version() != records["hstrat_version"]:
        log_once_in_a_row(
            f"""specimen_from_records version mismatch, record is version {
                records['hstrat_version']
            } and software is version {
                get_hstrat_version()
            }"""
        )

    policy = policy_from_record(records["policy"])
    differentia = numpy_fromiter_polyfill(
        unpack_differentiae_str(
            records["differentiae"],
            differentia_bit_width=records["differentia_bit_width"],
        ),
        dtype=np.min_scalar_type(2 ** records["differentia_bit_width"] - 1),
    )
    ranks = numpy_fromiter_polyfill(
        policy.IterRetainedRanks(records["num_strata_deposited"]),
        dtype=np.min_scalar_type(records["num_strata_deposited"] - 1),
    )
    return HereditaryStratigraphicSpecimen(
        pd.Series(data=differentia, index=ranks),
        records["differentia_bit_width"],
    )



import unittest

from hstrat._auxiliary_lib import alifestd_parse_ancestor_id


class TestOmitLast(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert alifestd_parse_ancestor_id("[]") is None
        assert alifestd_parse_ancestor_id("[None]") is None
        assert alifestd_parse_ancestor_id("[none]") is None
        assert alifestd_parse_ancestor_id("[NONE]") is None

    def test_singleton(self):
        assert alifestd_parse_ancestor_id("[1]") == 1
        assert alifestd_parse_ancestor_id("[12]") == 12
        assert alifestd_parse_ancestor_id("[0]") == 0
        assert alifestd_parse_ancestor_id("[1289]") == 1289


if __name__ == "__main__":
    unittest.main()



"""Logic to control stratum retention within HereditaryStratigraphicColumn.

Stratum retention policies specify which strata ranks that should be retained ----- and which should be purged --- when the nth stratum is deposited.
Stratum retention algorithms can be parameterized to yield a policy.
"""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import math
import typing

from ._iter_target_recencies import iter_target_recencies


def iter_target_ranks(
    degree: int,
    num_strata_deposited: int,
) -> typing.Iterator[int]:
    """Yield strata ranks for each exponentially-spaced coverage target
    `pow` in ascending order."""
    for target_recency in iter_target_recencies(degree, num_strata_deposited):
        recency_cutoff = target_recency
        rank_cutoff = max(
            num_strata_deposited - int(math.ceil(recency_cutoff)),
            0,
        )
        if num_strata_deposited == 0:
            assert rank_cutoff == 0
        else:
            assert 0 <= rank_cutoff <= num_strata_deposited - 1
        yield rank_cutoff



"""Eagerly utilize fixed stratum storage capacity `size_curb` to minimize
recency-proportional MRCA uncertainty.

The curbed recency-proportional resolution policy ensures that retained stratum
count respects a user-specified size cap. Recency-proportional MRCA resolution
is provided at finest-possible granularity given space constraint with graceful
degradation of granularity occuring as deposition history grows.

This policy works by splicing together successively-sparser
`recency_proportional_resolution_algo` parameterizations then --- after the
stratum retention count upper bound of the sparsest
`recency_proportional_resolution_algo` exceeds storage capacity --- a
permanently fixed parameterization of the `geometric_seq_nth root` algorithm.

While

    num_depositions < 2 ** (size_curb // 3) // 2,

`recency_proportional_resolution_algo` policies apply with resolution

    size_curb // ceil(log2(num_depisitons)) - 1.

For

    num_depisitons > 2 ** (size_curb // 3) // 2,

a `geometric_sequence_nth_root_algorithm` policy with degree

    (size_curb - 2) // 6

and interspersal 2 applies. To ensure availability of ranks required by the
`geometric_sequence_nth_root_algorithm`, the transition between algorithms
occurs exactly before recency-proportional resolution 1 would apply.

Because strata retained by each policy supersets
strata retained by all subsequent retention policies, uncertainty bounds for
all policies apply within their respective domains. (I.e., all strata expected
by each policy are available when that policy activates.)

Although the geometric sequence nth root algorithm also guarantees constant
space complexity with recency-proportional MRCA uncertainty, the curbed recency-
proportional algorithm makes fuller (i.e., more aggressive) use of available
space during early depositions.

The curbed recency-proportional algorithm also compares favorably to the
tapered geometric sequence nth root algorithm during early depositions.
Although the tapered nth root algorithm makes the fullest (i.e., most
aggressive) use of available space --- upper bound capacity is always perfectly
utilized --- the used space is less effective at minimizing worst-case recency-
proportional resolution than the curbed algorithm. (I.e., the tapered nth root
algorithm provides patchier, less systematic coverage.)

Notes
-----
Only size caps of eight strata and above are guaranteed respected.
(Initialization of smaller size caps will warn.)

See Also
--------
recency_proportional_resolution_algo:
    Provides log space complexity and recency-proportional strata spacing.
geom_seq_nth_root_algo:
    Provides constant space complexity and MRCA rank estimate uncertainty
    bounded under s * (1 - n^(-1/k)) where n is the number of strata deposited
    and s is the true number of ranks deposited since the MRCA.
geom_seq_nth_root_tapered_algo:
    Perfectly space-filling variant of geom_seq_nth_root_algo.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub
